const char *gen_mul_mfvf_vf_rcsid = "$Header: /u/krste/realsys/phipac/mul_mfvf_vf/RCS/gen_mul_mfvf_vf.c,v 1.5 1995/04/26 20:38:58 krste Exp davidj $";
/*
    Program that generates matrix-vector multiply routines.

    Takes two arguments: n_rows n_cols

    n_rows sets number of matrix rows to calculate simultaneously (default 4).
    n_cols sets number of columns unrolled in inner loop (default 1).
*/

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

static size_t n_rows = 4;
static size_t n_cols = 1;

static const char* prog_name;

static void
error(const char *msg)
{
    fprintf(stderr, "%s: %s\n", prog_name, msg);
    exit(1);
}


static size_t
parse_size_t(const char*const s)
{
    size_t len = strlen(s);
    char *ptr;
    const size_t res = (size_t) strtol(s, &ptr, 0);

    if (ptr != (s+len))
        error("Argument not an integer.");

    if (errno==ERANGE)
        error("Integer argument out of range.");

    return res;
}

static void
gen_header(void)
{
    printf("/* Code generated by %s */\n", prog_name);
    printf("/* Version: $Header: /u/krste/realsys/phipac/mul_mfvf_vf/RCS/gen_mul_mfvf_vf.c,v 1.5 1995/04/26 20:38:58 krste Exp davidj $ */\n");
    printf("/* n_rows=%ld n_cols=%ld */\n", n_rows, n_cols);

    printf("\n"
           "#include \"fltvec.h\"\n"
           "\n");

    printf("#ifdef FUNC_php_mul_mfvf_vf\n"
           "#define FP_TYPE float\n"
           "#define ROUT_NAME php_mul_mfvf_vf\n"
           "#define DOT_NAME php_mulsum_vfvf_f\n"
           "#endif\n");
    printf("\n");
    printf("#ifdef FUNC_php_mul_mdvd_vd\n"
           "#define FP_TYPE double\n"
           "#define ROUT_NAME php_mul_mdvd_vd\n"
           "#define DOT_NAME php_mulsum_vdvd_d\n"
           "#endif\n");
    printf("\n");

    /* This should really be in fltvec.h already. */
    printf("extern FP_TYPE\n"
           "DOT_NAME(size_t,const FP_TYPE*,const FP_TYPE*);\n"
           "\n");

    printf("#define RB (%ld)\n", n_rows);
    printf("#define RC (%ld)\n", n_cols);
    printf("\n");
}

/* Returns 1 if argument a power of 2, 0 otherwise. */
static int
powerof2(size_t x)
{
    return !(x & (x-1));
}

static void
gen_row_blocked(void)
{
    size_t r;
    size_t c;

    /* Function proper starts. */

    printf("void\n"
           "ROUT_NAME(const size_t rows, const size_t cols, const size_t ld,\n"
           "          const FP_TYPE* mat, const FP_TYPE* vec, FP_TYPE* res)\n");
    printf("{\n");

    printf("\tconst FP_TYPE *rowp = mat;\n"
           "\tsize_t i;\n"
           "\n");

    /* Not all compilers do a good job of optimizing away mod operation. */
    printf("\tconst size_t last_block_row = ");
    if (powerof2(n_rows))
        printf("rows & ~(RB-1);\n");
    else
        printf("rows - (rows %% RB);\n");

    /* Hoist this constant out of loop. */
    printf("\tconst FP_TYPE* end_inp = vec + cols;\n"
           "\n");

    if (n_cols > 1)
    {
        /* Only needed if unrolling columns. */
        /* Hoist this value out of loop. */
        printf("\tconst FP_TYPE* end_col_unrollp = vec + ");
        if (powerof2(n_cols))
            printf("(cols & ~(RC-1));\n");
        else
            printf("(cols - (cols %% RC));\n");
    }

    /* Outer loop computes sums for n_rows rows at one time. */

    printf("\tfor (i = 0 ; i != last_block_row; i += RB)\n");
    printf("\t{\n");

    for (r=0; r<n_rows; r++)
    {
        printf("\t\tFP_TYPE rsum%ld = 0.0;\n", r);
        if (r==0)
            printf("\t\tconst FP_TYPE* rp%ld = rowp;\n", r);
        else
            printf("\t\tconst FP_TYPE* rp%ld = rp%ld + ld;\n", r, r-1);
    }
    printf("\n");

    printf("\t\tconst FP_TYPE* inp;\n");

    if (n_cols > 1)
    {
        /* Unroll columns. */
        printf("\t\tfor (inp = vec; inp != end_col_unrollp; inp += RC)\n"
               "\t\t{\n");

        for (c=0; c<n_cols; c++)
            printf("\t\t\tconst FP_TYPE velem%ld = inp[%ld];\n", c, c);

        for (c=0; c<n_cols; c++)
        {
            printf("\n");
            for (r=0; r<n_rows; r++)
                printf("\t\t\trsum%ld += velem%ld * rp%ld[%ld];\n", r,c,r,c);
        }

        for (r=0; r<n_rows; r++)
            printf("\t\t\trp%ld += RC;\n", r);
        printf("\t\t}\n");
        printf("\n");

        /* Finish off remaining columns. */
        printf("\t\tfor ( ; inp != end_inp; inp++)\n"
               "\t\t{\n"
               "\t\t\tconst FP_TYPE velem = *inp;\n"
               "\n");
        for (r=0; r<n_rows; r++)
            printf("\t\t\trsum%ld += velem * (*rp%ld++);\n", r, r);
        printf("\t\t}\n");
        printf("\n");
    }
    else
    {
        /* Simple loop over columns. */
        printf("\t\tinp = vec;\n");
        printf("\t\twhile (inp != end_inp)\n"
               "\t\t{\n"
               "\t\t\tconst FP_TYPE velem = *inp++;\n"
               "\n");
        for (r=0; r<n_rows; r++)
            printf("\t\t\trsum%ld += velem * (*rp%ld++);\n", r, r);
        printf("\t\t}\n");
    }

    for (r=0; r<n_rows; r++)
        printf("\t\tres[%ld] = rsum%ld;\n", r, r);
    printf("\n");
    printf("\t\tres += RB;\n");
    printf("\n");
    printf("\t\trowp = rp%ld;\n", n_rows-1);

    printf("\t}\n");
    printf("\n");

    /* Finish off remaining rows using fast dot-product. */

    printf("\tfor ( ; i<rows; i++)\n"
           "\t{\n"
           "\t\t*res++ = DOT_NAME(cols, rowp, vec);\n"
           "\t\trowp += ld;\n"
           "\t}\n");

    printf("}\n");
}

static void
gen_with_dot(void)
{
    printf("void\n"
           "ROUT_NAME(const size_t rows, const size_t cols, const size_t ld,\n"
           "          const FP_TYPE* mat, const FP_TYPE* vec, FP_TYPE* res)\n");
    printf("{\n");
    printf("\tsize_t i;\n");

    printf("\tfor (i=0; i<rows; i++)\n"
           "\t{\n"
           "\t\t*res++ = DOT_NAME(cols, mat, vec);\n"
           "\t\tmat += ld;\n"
           "\t}\n");

    printf("}\n");
}

static void
gen_code(void)
{
    gen_header();
    if (n_rows > 1)
        gen_row_blocked();
    else
        gen_with_dot();
}

int
main(int argc, char *argv[])
{
    prog_name = argv[0];

    if (argc > 3)
        error("Too many arguments.");

    if (argc > 1)
        n_rows = parse_size_t(argv[1]);

    if (argc > 2)
        n_cols = parse_size_t(argv[2]);

    gen_code();

    return 0;
}
